#pragma once

/// @brief Top-level namespace
namespace dolfinx
{
}

// DOLFINx interface

#include <dolfinx/common/dolfinx_common.h>
#include <dolfinx/fem/dolfinx_fem.h>
#include <dolfinx/geometry/dolfinx_geometry.h>
#include <dolfinx/graph/dolfinx_graph.h>
#include <dolfinx/io/dolfinx_io.h>
#include <dolfinx/la/dolfinx_la.h>
#include <dolfinx/mesh/dolfinx_mesh.h>
#include <dolfinx/nls/dolfinx_nls.h>
#include <dolfinx/refinement/dolfinx_refinement.h>
