# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import datetime
import os
import sys

import mpi4py

import numpy as np
import packaging

import basix
import dolfinx
import ffcx
import ufl

sys.path.insert(0, os.path.abspath("."))
import jupytext_process  # isort:skip


myst_heading_anchors = 3

jupytext_process.process()

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "myst_parser",
    "sphinx_codeautolink",
    "sphinx.ext.intersphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "DOLFINx"
now = datetime.datetime.now()
date = now.date()
copyright = f"{date.year}, FEniCS Project"
author = "FEniCS Project"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = dolfinx.__version__
# The full version, including alpha/beta/rc tags.
release = dolfinx.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "DOLFINxdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "DOLFIN.tex", "DOLFINx Documentation", "FEniCS Project", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "dolfinx", "DOLFINx Documentation", [author], 1)]

autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
    "imported-members": True,
    "undoc-members": True,
}
autosummary_generate = True
autosummary_ignore_module_all = False
autoclass_content = "both"

codeautolink_concat_default = True

intersphinx_resolve_self = "dolfinx"

# Could be reimplemented using packaging.version
basix_version = "main" if "dev0" in basix.__version__ else "v" + basix.__version__
ffcx_version = "main" if "dev0" in ffcx.__version__ else "v" + ffcx.__version__
ufl_version = "main" if "dev0" in ufl.__version__ else ufl.__version__

numpy_version = packaging.version.parse(np.__version__)
numpy_doc_version = f"{numpy_version.major}.{numpy_version.minor}"

# Note that as of late 2025 pyvista and petsc4py only have docs for the latest
# releases.
intersphinx_mapping = {
    "petsc4py": (
        "https://petsc.org/release/petsc4py",
        None,
    ),
    "numpy": (
        f"https://numpy.org/doc/{numpy_doc_version}",
        None,
    ),
    "pyvista": (
        "https://docs.pyvista.org",
        None,
    ),
    "mpi4py": (
        f"https://mpi4py.readthedocs.io/en/{mpi4py.__version__}",
        None,
    ),
    "basix": (
        f"https://docs.fenicsproject.org/basix/{basix_version}/python",
        None,
    ),
    "ffcx": (
        f"https://docs.fenicsproject.org/ffcx/{ffcx_version}",
        None,
    ),
    "ufl": (
        f"https://docs.fenicsproject.org/ufl/{ufl_version}",
        None,
    ),
}

napoleon_google_docstring = True
napoleon_use_admonition_for_notes = False

myst_enable_extensions = [
    "dollarmath",
]


def skip_member(app, what, name, obj, skip, opts):
    # Skip @entries from nanobind enums
    if name == "@entries":
        return True
    else:
        return skip


def setup(app):
    app.connect("autodoc-skip-member", skip_member)
