#pragma once

/// @brief Graph data structures and algorithms.
///
/// Data structures for building and representing graphs, and algorithms
/// on graphs, e.g., re-ordering and partitioning.
namespace dolfinx::graph
{
}

// DOLFINx graph interface

#include <dolfinx/graph/partition.h>
