#pragma once

/// @brief Mesh refinement algorithms.
///
/// Methods for refining meshes uniformly, or with markers, using edge
/// bisection.
namespace dolfinx::refinement
{
}

// DOLFINx refinement interface

#include <dolfinx/refinement/interval.h>
#include <dolfinx/refinement/refine.h>
#include <dolfinx/refinement/uniform.h>
